theory SSS
  imports Main
begin

declare [[smt_oracle]]

(* <DEFINITIONS> *)

type_synonym ('a) state = "'a"

datatype 'a transition = Transition (src: "'a state")  (dst: "'a state") (weight: "nat")

type_synonym 'a operator = "'a transition set"

datatype 'a ts = TS (states: "'a state set") (transitions: "'a transition set") (init: "'a state") (goals: "'a state set") (*(operators: "'a operator set")*)

type_synonym 'a path = "'a operator list"

definition valid_operator :: "'a ts \<Rightarrow> 'a operator \<Rightarrow> bool" where
  "valid_operator T op \<equiv> (op \<noteq> {} \<and> op \<subseteq> transitions T \<and> \<not>(\<exists> t \<in> op. \<exists> t' \<in> op. t \<noteq> t' \<and> src t = src t') \<and> (\<forall> t \<in> op. \<forall> t' \<in> op. weight t = weight t'))"


definition cost :: "'a operator \<Rightarrow> nat" where
"cost op = (SOME n :: nat. \<exists> t \<in> op. weight t = n)"


definition applicable :: "'a operator \<Rightarrow> 'a state \<Rightarrow> bool" where
"applicable op s \<equiv> \<exists> t \<in> op. src t = s"


definition effect :: "'a ts \<Rightarrow> 'a operator \<Rightarrow> 'a state \<Rightarrow> 'a state" where
"effect T op s \<equiv> (SOME s' :: 'a state. s' \<in> states T \<and> (\<exists> t \<in> op. (s = src t \<and> s' = dst t)))"


definition enables :: "'a ts \<Rightarrow> 'a operator \<Rightarrow> 'a operator \<Rightarrow> bool" where
"enables T op op' \<equiv> \<forall> s \<in> states T. (applicable op s \<and> applicable op' s) \<longrightarrow> applicable op' (effect T op s)"


definition non_conflict :: "'a ts \<Rightarrow> 'a operator \<Rightarrow> 'a operator \<Rightarrow> bool" where
"non_conflict T op op' \<equiv> \<forall> s \<in> states T. (applicable op s \<and> applicable op' s \<and> applicable op (effect T op' s) \<and> applicable op' (effect T op s)) \<longrightarrow> effect T op (effect T op' s) = effect T op' (effect T op s)"


definition independent :: "'a ts \<Rightarrow> 'a operator \<Rightarrow> 'a operator \<Rightarrow> bool" where
"independent T op op' \<equiv> enables T op op' \<and> enables T op' op \<and> non_conflict T op op'"


definition valid_ts::"'a ts \<Rightarrow> bool" where
 "valid_ts T \<equiv> init T \<in> states T \<and> goals T \<subseteq> states T \<and>
(\<forall> trans \<in> transitions T. src trans \<in> states T \<and> dst trans \<in> states T)"


fun comes_before :: "'a \<Rightarrow> 'a \<Rightarrow> 'a list \<Rightarrow> bool" where
"comes_before s s' [] = False" |
"comes_before s s' (x#xs) = (if (s = x \<and> s \<noteq> s') then True else if s' = x then False else comes_before s s' xs)"


fun valid_path :: "'a ts \<Rightarrow> 'a path \<Rightarrow> 'a state \<Rightarrow> bool" where
"valid_path T [] s = True" |
"valid_path T (op#ops) s = (applicable op s \<and> valid_operator T op \<and> valid_path T ops (effect T op s))"


fun is_path_from_to :: "'a ts \<Rightarrow> 'a path \<Rightarrow> 'a state \<Rightarrow> 'a state \<Rightarrow> bool" where
"is_path_from_to T [] s s' = (s = s')" |
"is_path_from_to T (op#ops) s s' = (applicable op s \<and> is_path_from_to T ops (effect T op s) s')"


fun path_effect :: "'a ts \<Rightarrow> 'a path \<Rightarrow> 'a state \<Rightarrow> 'a state" where
"path_effect T [] s = s" |
"path_effect T (op # ops) s = path_effect T ops (effect T op s)"


lemma path_effect_0: "valid_path T pi s \<Longrightarrow> is_path_from_to T pi s s' \<Longrightarrow> path_effect T pi s = s'" 
proof(induct pi arbitrary: s s')
  case Nil
  then show ?case by simp
next
  case (Cons a pi)
  then show ?case by simp
qed


lemma path_effect_1: "valid_path T pi s \<Longrightarrow> is_path_from_to T pi s (path_effect T pi s)"
proof (induct pi arbitrary: s)
case Nil
  then show ?case by simp
next
  case (Cons a pi)
  then have "applicable a s \<Longrightarrow> valid_path T pi s \<Longrightarrow> is_path_from_to T (a#pi) s (path_effect T (a#pi) s) \<Longrightarrow> ?case " by simp
  from this have "\<lbrakk>applicable a s; valid_path T pi s; is_path_from_to T pi (effect T a s) (path_effect T (a#pi) s)\<rbrakk> \<Longrightarrow> ?case" by simp
  from this show ?case using Cons.hyps Cons.prems by auto
qed


definition necessary_enabling_set :: "'a ts \<Rightarrow> 'a operator set \<Rightarrow>'a state \<Rightarrow> 'a operator \<Rightarrow> bool" where
"necessary_enabling_set T N s op \<equiv>
 (\<forall> op' \<in> N. valid_operator T op') \<and>
 (\<forall> goal \<in> goals T. \<forall> pi. (\<not> applicable op s \<and> valid_path T pi s \<and> is_path_from_to T pi s goal \<and> op \<in> set pi) \<longrightarrow> (\<exists>op' \<in> N. op' \<in> set pi \<and> comes_before op' op pi))"


definition disjunctive_landmark :: "'a ts \<Rightarrow> 'a operator set \<Rightarrow> 'a state \<Rightarrow> bool" where
"disjunctive_landmark T L s \<equiv>
 s \<in> states T \<and>
 (\<forall> op \<in> L. valid_operator T op) \<and>
 (\<forall> pi ::'a path. \<forall> goal \<in> goals T. (valid_path T pi s \<and> is_path_from_to T pi s goal) \<longrightarrow> (\<exists>op' \<in> set pi. op' \<in> L ))"


definition active :: "'a ts \<Rightarrow> 'a state \<Rightarrow> 'a state \<Rightarrow> bool" where
"active T s s' \<equiv>
(\<exists> pi. valid_path T pi s \<and> is_path_from_to T pi s s') \<and>
(\<exists> goal \<in> goals T. \<exists> pi'. valid_path T pi' s' \<and> is_path_from_to T pi' s' goal)"


definition plan :: "'a ts \<Rightarrow> 'a state \<Rightarrow> 'a path \<Rightarrow> bool" where
"plan T s pi \<equiv> (s \<in> states T) \<and> (valid_path T pi s) \<and> (\<exists> goal \<in> goals T. is_path_from_to T pi s goal)"


fun path_cost :: "'a path \<Rightarrow> nat" where
"path_cost [] = 0" |
"path_cost (op # ops) = cost op + path_cost ops"


definition optimal_plan :: "'a ts \<Rightarrow>'a state \<Rightarrow> 'a path \<Rightarrow> bool" where
"optimal_plan T s pi \<equiv> plan T s pi \<and> (path_cost pi = Min{path_cost pi' | pi'. plan T s pi'})"


fun path_independence :: "'a ts \<Rightarrow> 'a path \<Rightarrow> 'a operator \<Rightarrow> bool" where
"path_independence T [] op = True" |
"path_independence T (x#xs) op = (independent T x op \<and> path_independence T xs op)"


fun first_occurrence :: "'a path \<Rightarrow> 'a operator set \<Rightarrow> 'a operator" where
"first_occurrence [] S = {}" |
"first_occurrence (op # ops) S = (if op \<in> S then op else first_occurrence ops S)"


fun swap_with_previous :: "'a list \<Rightarrow> 'a \<Rightarrow> 'a list" where
"swap_with_previous [] _ = []" |
"swap_with_previous [x] op = [x]" |
"swap_with_previous (x#y#xs) op = (if x = op then (x#y#xs) else if y = op then (y#x#xs) else (x # swap_with_previous (y#xs) op))"


fun pull :: "'a list \<Rightarrow> nat \<Rightarrow> 'a \<Rightarrow> 'a list" where
"pull xs 0 op = xs" |
"pull xs (Suc n) op = pull (swap_with_previous xs op) n op"


definition strong_stubborn_set :: "'a ts \<Rightarrow> 'a operator set \<Rightarrow> 'a state \<Rightarrow> bool" where
"strong_stubborn_set T SSS s \<equiv>
(\<forall> op \<in> SSS. valid_operator T op) \<and>
(\<forall> op' \<in> SSS. \<not> applicable op' s \<longrightarrow> necessary_enabling_set T SSS s op') \<and>
(\<forall> op \<in> SSS. (applicable op s \<longrightarrow> (\<forall> op'::'a operator. (valid_operator T op' \<and> \<not> independent T op op') \<longrightarrow> (op' \<in> SSS)))) \<and>
(\<exists> L :: 'a operator set. disjunctive_landmark T L s \<and> L \<subseteq> SSS)"


fun repeat :: "('a list \<Rightarrow> 'a \<Rightarrow> 'a list) \<Rightarrow> nat \<Rightarrow> 'a list \<Rightarrow> 'a \<Rightarrow> 'a list" where
"repeat f 0 pi op = pi" |
"repeat f (Suc n) pi op = f (repeat f n pi op) op"

(* </DEFINITIONS> *)

(* <ELEMENTARY LEMMAS> *)

lemma "non_conflict T op op' \<Longrightarrow> non_conflict T op' op"
  by (metis non_conflict_def)


lemma independent_symmetric: "independent T op op' \<Longrightarrow> independent T op' op" using independent_def
  by (metis non_conflict_def)


lemma comes_before_itself: "\<not> comes_before a a pi"
proof (induct pi)
  case Nil
  then show ?case by auto
next
  case (Cons x xs)
  then show ?case by simp
qed


lemma notApplicableInNecessaryEnablingSet:
  fixes T :: "'a ts" and pi :: "'a path" and s :: "'a state" and op :: "'a operator" and N :: "'a operator set" and goal :: "'a state"
  assumes necessaryEnablingSet: "necessary_enabling_set T N s op" and
          goalState: "goal \<in> goals T" and
          validPath: "valid_path T pi s" and
          pathToGoal: "is_path_from_to T pi s goal" and
          opInPi: "op \<in> set pi" and
          notApplicable: "\<not> applicable op s"
        shows "\<exists> op' \<in> N. comes_before op' op pi \<and> op' \<in> set pi"
  by (meson goalState necessaryEnablingSet necessary_enabling_set_def notApplicable opInPi pathToGoal validPath)


lemma operators_of_SSS_are_valid : "\<lbrakk>strong_stubborn_set T SSS s ; op \<in> SSS\<rbrakk> \<Longrightarrow> valid_operator T op"
  by (simp add: strong_stubborn_set_def)


lemma not_applicable_in_SSS : "\<lbrakk>strong_stubborn_set T SSS s ; op \<in> SSS ; \<not> applicable op s\<rbrakk> \<Longrightarrow> necessary_enabling_set T SSS s op"
  by (simp add: strong_stubborn_set_def)


lemma applicable_in_SSS : "\<lbrakk>strong_stubborn_set T SSS s ; op \<in> SSS; applicable op s; valid_operator T op' \<and> \<not> independent T op op'\<rbrakk> \<Longrightarrow> op' \<in> SSS" 
  by (meson strong_stubborn_set_def)


(* If L is a disjunctive landmark and pi is an optimal plan, then L \<inter> set pi \<noteq> {} *)
lemma lemma1:
"(valid_ts T \<and> optimal_plan T s pi \<and> pi \<noteq> [] \<and> disjunctive_landmark T L s) \<Longrightarrow> (L \<inter> (set pi) \<noteq> {})"
  unfolding optimal_plan_def plan_def disjunctive_landmark_def by blast 


(* If L is a set and pi a path s.t. their intersection is not empty, then the element with the smallest index in pi that is also in L is an element of L \<inter> set pi; in particular, it exists *)
lemma lemma2:
  fixes L :: "'a operator set" and pi :: "'a path"
  shows "(L \<inter> set pi \<noteq> {} \<and> (\<forall> op \<in> L \<inter> set pi. op \<noteq> {})) \<Longrightarrow> (pi = [] \<longrightarrow> first_occurrence pi L = {}) \<and> (pi \<noteq> [] \<longrightarrow>first_occurrence pi L \<in> L \<inter> set pi)"
proof (induct pi)
  case Nil
  from this show ?case by auto
next
  case (Cons x xs)

  from this have ONE: "(x \<notin> L \<and> xs = []) \<Longrightarrow> ?case" by simp
  from local.Cons have TWO: "(x \<notin> L \<and> xs \<noteq> []) \<Longrightarrow> ?case" by simp
  from local.Cons have THREE: "x \<in> L \<Longrightarrow> ?case" by simp

  from ONE and TWO and THREE show ?case by auto
qed


(* if pi is a valid path then the elements of set(pi) are valid operators *)
lemma lemma3:
(* https://stackoverflow.com/questions/60684922 *)
  fixes pi :: "'a path" and T :: "'a ts"
  shows "valid_path T pi s \<Longrightarrow> \<forall> op \<in> set pi. valid_operator T op"
proof (induct pi arbitrary: s)
  case Nil
  then show ?case by auto
next
  case (Cons x xs)
  from this show ?case using \<open>valid_path T xs (effect T x s) \<Longrightarrow> \<forall>op\<in>set xs. valid_operator T op\<close> by auto
qed


(* if pi is an optimal plan, then the elements of set(pi) are valid operators *)
lemma lemma4:
  fixes pi :: "'a path" and s :: "'a state" and T :: "'a ts"
  assumes optimalPlan: "optimal_plan T s pi" and sInT: "s \<in> states T" and optimalPlan: "optimal_plan T s pi"
  shows "\<forall> op \<in> set pi. valid_operator T op"
proof -
  from optimalPlan have "plan T s pi" unfolding optimal_plan_def by simp
  from this and plan_def have "valid_path T pi s" unfolding plan_def by simp
  from this and lemma3 show ?thesis by fastforce
qed


(* if
a comes before b and
a, b \<in> S and
a, b \<in> set pi and
a \<noteq> b
then
first_occurrence pi S cannot be equal to b (since a comes before b) *)
lemma lemma5: "(comes_before a b pi \<and> a \<in> S \<and> b \<in> S \<and> a \<in> set pi \<and> b \<in> set pi \<and> a \<noteq> b) \<Longrightarrow> first_occurrence pi S \<noteq> b"
proof (induct pi)
  case Nil
  then show ?case by auto
next
  case (Cons x xs)
  from local.Cons have 1: "comes_before a b (x#xs) \<and> a = x \<and> b = x \<Longrightarrow> ?case" by auto
  from local.Cons have 2: "comes_before a b (x#xs) \<and> a = x \<and> b \<noteq> x \<Longrightarrow> ?case" by auto
  from local.Cons have 3: "comes_before a b (x#xs) \<and> a \<noteq> x \<and> b = x \<Longrightarrow> ?case" by auto
  from local.Cons have 4: "comes_before a b (x#xs) \<and> a \<noteq> x \<and> b \<noteq> x \<Longrightarrow> ?case" by auto
  from 1 and 2 and 3 and 4 show ?case using Cons.prems by fastforce
qed


(* It is impossible that all of the following hold at the same time:
op, op' \<in> SSS and
op, op' \<in> set pi and
op = first_ocurrace pi SSS and
op' comes before op in pi *)
lemma lemma6:
  fixes SSS :: "'a operator set" and op :: "'a operator" and op' :: "'a operator" and pi :: "'a path" and T :: "'a ts"
  assumes 1: "op' \<in> SSS" and
          2: "op \<in> SSS" and
          3: "op \<in> set pi" and
          4: "op' \<in> set pi" and
          5: "op = first_occurrence pi SSS" and
          6: "comes_before op' op pi"
  shows "False"
proof -
  from 1 and 2 and 3 and 4 and 5 and 6 and lemma5 have "op = op'" by auto
  from this and 6 have "comes_before op' op' pi" by auto
  from this show ?thesis by (simp add: comes_before_itself)
qed


(* If op is s.t. it is the first occurrence in pi and SSS and it is applicable in s, then it is independent from all of its previous operators in pi*)
lemma lemma7:
  fixes T :: "'a ts" and s :: "'a state" and SSS :: "'a operator set" and op :: "'a operator" and pi :: "'a path" and op' :: "'a operator"
  assumes validTS: "valid_ts T" and
          strStbSet : "strong_stubborn_set T SSS s" and
          sInT : "s \<in> states T" and
          optimal: "optimal_plan T s pi \<and> pi \<noteq> []" and
          first: "op = first_occurrence pi SSS" and
          opInSSSandPi:  "op \<in> SSS \<inter> set pi" and
          op'InPi: "op' \<in> set pi" and
          op'before: "comes_before op' op pi" and
          applicable: "applicable op s" and
          dep: "\<not> independent T op op'"
        shows "False"
proof -
  from optimal and lemma4 and sInT have "valid_operator T op'" using op'InPi by fastforce
  from this and applicable and opInSSSandPi and strStbSet and strong_stubborn_set_def have "op' \<in> SSS" by (meson Int_iff applicable_in_SSS dep)
  from this show ?thesis using first lemma6 op'InPi op'before opInSSSandPi by auto
qed

(* </ELEMENTARY LEMMAS *)

(* <SWAP WITH PREVIOUS & PULL> *)

lemma len_swap_with_prev: "length (swap_with_previous xs op) = length xs"
proof (induct xs)
  case Nil
  then show ?case by simp
next
  case (Cons x xs)
  then have one: "x = op \<Longrightarrow> ?case" by (metis neq_Nil_conv swap_with_previous.simps(2) swap_with_previous.simps(3))

  also have two: "x \<noteq> op \<and> xs = [] \<Longrightarrow> ?case" by simp

  also have "x \<noteq> op \<and> hd xs = op \<and> xs \<noteq> [] \<Longrightarrow> swap_with_previous (x#xs) op = op # x # tl xs" by (metis list.exhaust_sel swap_with_previous.simps(3))
  from this have 1: "x \<noteq> op \<and> hd xs = op \<and> xs \<noteq> [] \<Longrightarrow> ?case" by auto
  from local.Cons have "x \<noteq> op \<and> hd xs \<noteq> op \<and> xs \<noteq> [] \<Longrightarrow> swap_with_previous (x#xs) op = x # swap_with_previous xs op"  by (metis list.exhaust_sel swap_with_previous.simps(3)) 
  from this have 2: "x \<noteq> op \<and> hd xs \<noteq> op \<and> xs \<noteq> [] \<Longrightarrow> ?case" by (simp add: Cons.hyps)
  from 1 and 2 have three:"x \<noteq> op \<and> xs \<noteq> [] \<Longrightarrow> ?case" by auto

  from one and two and three show ?case by blast
qed


lemma swap_head_with_prev: "swap_with_previous pi (hd pi) = pi"
  by (metis (no_types, hide_lams) len_swap_with_prev length_0_conv list.sel(1) neq_Nil_conv swap_with_previous.simps(2) swap_with_previous.simps(3))


lemma pull_head_to_head_does_nothing: "pull pi n (hd pi) = pi"
proof (induct n arbitrary: pi)
  case 0
  then show ?case by simp
next
  case (Suc n)
  then have "pull pi (Suc n) (hd pi) = pull (swap_with_previous pi (hd pi)) n (hd pi) "by simp
  from this have "pull pi (Suc n) (hd pi) = pull pi n (hd pi)" by (simp add: swap_head_with_prev)
  from this show ?case by (simp add: Suc.hyps)
qed


lemma pull_and_swap_commute: "pull (swap_with_previous pi op) n op = swap_with_previous (pull pi n op) op"
proof (induct n arbitrary: pi)
  case 0
  then show ?case by simp
next
  case (Suc n)
  then show ?case by simp
qed


lemma pull_eq_repeat_swap: "repeat swap_with_previous n pi op  = pull pi n op"
proof (induct n arbitrary: pi)
  case 0
  then show ?case by simp
next
  case (Suc n)
  then have 1: "repeat swap_with_previous (Suc n) pi op = swap_with_previous (repeat swap_with_previous n pi op) op" by simp
  also have 2: "pull pi (Suc n) op = pull (swap_with_previous pi op) n op" by simp
  from 1 and 2 and local.Suc show ?case by (simp add: pull_and_swap_commute)
qed


lemma head_concat_swap: "op \<in> set xs \<Longrightarrow> x \<noteq> op \<Longrightarrow> op \<noteq> hd xs \<Longrightarrow> x # (swap_with_previous xs op) = swap_with_previous (x#xs) op"
proof (induct xs)
  case Nil
  then show ?case by simp
next
  case (Cons x xs)
  then show ?case by auto
qed


lemma set_swap_eq:"set xs = set (swap_with_previous xs op)"
proof (induct xs)
  case Nil
  then show ?case by simp
next
  case (Cons x xs)
  then have "xs = [] \<Longrightarrow> ?case" by simp
  from local.Cons have 0: "x = op \<and> xs \<noteq> [] \<Longrightarrow> ?case"
proof -
  assume "x = op \<and> xs \<noteq> []"
  then show ?thesis
    by (metis list.sel(1) swap_head_with_prev)
qed
  from local.Cons have 1: "x \<noteq> op \<and> xs \<noteq> [] \<and> op \<notin> set xs \<Longrightarrow> ?case" by (metis hd_in_set list.exhaust_sel list.simps(15) swap_with_previous.simps(3))
  from local.Cons have 2:  "x \<noteq> op \<and> xs \<noteq> [] \<and> op \<in> set xs \<Longrightarrow> set (x # xs) \<subseteq> set (swap_with_previous (x # xs) op)" by (smt in_set_member list.exhaust_sel member_rec(1) subsetI swap_with_previous.simps(3))
  from local.Cons have 3:  "x \<noteq> op \<and> xs \<noteq> [] \<and> op \<in> set xs \<Longrightarrow> set (x # xs) \<supseteq> set (swap_with_previous (x # xs) op)" by (smt in_set_member list.exhaust_sel member_rec(1) subsetI swap_with_previous.simps(3))
  from local.Cons have 4: "x = op \<and> xs \<noteq> [] \<Longrightarrow> ?case" using 0 by auto
  from local.Cons have 5: "xs = [] \<Longrightarrow> ?case" by simp
  from 1 and 2 and 3 and 4 and 5 show ?case by (meson eq_iff)
qed


lemma list_1: "length pi > 1 \<Longrightarrow> pi = butlast pi @ [last pi]" 
  by (metis append_butlast_last_id gr_implies_not0 length_0_conv)


lemma list_2: "op \<notin> set pi \<Longrightarrow> op \<noteq> q \<Longrightarrow> swap_with_previous (pi @ [q, op]) op = pi @ [op, q]"
proof (induct pi)
  case Nil
  then show ?case by simp
next
  case (Cons a pi)
  then show ?case by (metis append_Cons head_concat_swap last.simps last_appendR last_in_set list.discI list.set_intros(1) list.set_intros(2) swap_head_with_prev)
qed


lemma list_3: "op \<notin> set pi \<Longrightarrow> op \<noteq> q \<Longrightarrow> op \<noteq> x \<Longrightarrow> swap_with_previous (x # pi @ [q, op]) op = x # pi @ [op, q]" by (metis append_Cons list_2 set_ConsD)


lemma list_4 : "length pi = 1 \<Longrightarrow> \<exists> q. pi = [q]"
  by (metis cancel_comm_monoid_add_class.diff_cancel length_greater_0_conv length_tl less_one list.exhaust_sel not_gr0)


lemma list_5: "op \<notin> set pi \<Longrightarrow> pi \<noteq> [] \<Longrightarrow> swap_with_previous (pi @ [op]) op = (butlast pi) @ [op, last pi]"
proof (induct pi)
  case Nil
  then show ?case by simp
next
  case (Cons x pi)

  moreover {assume "length pi = 1"
    from this obtain q where "pi = [q] \<and> q \<noteq> op" using calculation(2) list_4 by fastforce
    from this have "swap_with_previous ((x # pi) @ [op]) op = swap_with_previous [x, q, op] op" by auto
    from this have "swap_with_previous ((x # pi) @ [op]) op = [x, op, q]" using Cons.prems(1) \<open>pi = [q] \<and> q \<noteq> op\<close> by auto
    from this have ?case by (simp add: \<open>pi = [q] \<and> q \<noteq> op\<close>)
  }

  moreover {assume "length pi > 1"
    from this obtain q where "q = last pi \<and> q \<noteq> op" by (metis calculation(2) last_in_set length_0_conv less_numeral_extra(2) list.set_intros(2))
    from this have "swap_with_previous ((x # pi) @ [op]) op = swap_with_previous ((x # butlast pi @ [q]) @ [op]) op" using \<open>1 < length pi\<close> list_1 by fastforce
    from this have "swap_with_previous ((x # pi) @ [op]) op = swap_with_previous (x # butlast pi @ [q, op]) op" by auto
    from this have "swap_with_previous ((x # pi) @ [op]) op = x # butlast pi @ [op, q]" by (metis \<open>q = last pi \<and> q \<noteq> op\<close> calculation(2) in_set_butlastD list.set_intros(1) list.set_intros(2) list_3)
    from this have ?case using \<open>1 < length pi\<close> \<open>q = last pi \<and> q \<noteq> op\<close> by auto
  }

  ultimately show ?case using linorder_neqE_nat by fastforce
qed

lemma list_6: "op \<notin> set pi \<Longrightarrow> swap_with_previous pi op = pi" 
proof(induct pi)
  case Nil
  then show ?case by simp
next
  case (Cons a pi)
  then show ?case by (metis list.exhaust_sel list.set_intros(2) list.set_sel(1) swap_with_previous.simps(2) swap_with_previous.simps(3))
qed


lemma list_9: "xs \<noteq> [] \<Longrightarrow> op \<notin> set (x#xs) \<Longrightarrow> swap_with_previous ((x # xs) @ [op]@ ys) op = x # swap_with_previous (xs @ [op] @ ys) op"
proof(induct xs)
  case Nil
  then show ?case by simp
next
  case (Cons a xs)
  then show ?case by auto
qed


lemma list_10: "op \<notin> set xs \<Longrightarrow> xs \<noteq> [] \<Longrightarrow> swap_with_previous (xs @ [op] @ ys) op = (butlast xs) @ [op, last xs] @ ys"
proof (induct xs)
  case Nil
  then show ?case by simp
next
  case (Cons x xs)
  have "xs \<noteq> [] \<Longrightarrow> op \<notin> set(x#xs) \<Longrightarrow> swap_with_previous ((x#xs) @ [op] @ ys) op = x#swap_with_previous (xs @ [op] @ ys) op" by (meson list_9)
  from this have "op \<notin> set(x#xs) \<Longrightarrow> x#swap_with_previous (xs @ [op] @ ys) op = x # (butlast xs) @ [op, last xs] @ ys \<Longrightarrow> ?case" by simp
  from this show ?case using Cons.hyps Cons.prems(1) by auto
qed


lemma list_17: "set pi = set(pull pi n op)"
proof (induct n)
  case 0
  then show ?case by simp
next case (Suc n)
  then show ?case by (simp add: pull_and_swap_commute set_swap_eq)
qed


lemma pull_suc_n:"op \<in> set xs \<Longrightarrow> hd (pull xs n op) = op \<Longrightarrow> hd (pull (x#xs) (Suc n) op) = op"
proof(induct n arbitrary: xs)
  case 0
  then show ?case 
    by (metis empty_iff list.sel(1) neq_Nil_conv pull.simps(2) pull_eq_repeat_swap repeat.simps(1) set_empty2 swap_with_previous.simps(3))
next
  case (Suc n)
  have 1: "op \<in> set xs \<Longrightarrow> hd (pull xs (Suc n) op) = op \<Longrightarrow> hd (pull (swap_with_previous xs op) n op) = op" by auto
  have 2: "op \<in> set xs \<Longrightarrow> hd (pull (x#xs) (n+2) op) = op \<Longrightarrow> hd (pull (swap_with_previous (x#xs) op) (n+1) op) = op" by auto

  from 1 and 2 and local.Suc and pull_head_to_head_does_nothing and swap_with_previous.simps(3) and list.set_cases pull.simps(2) and list.sel(1)
  have 3: "op \<in> set xs \<Longrightarrow> hd (pull (swap_with_previous xs op) n op) = op \<Longrightarrow> hd (pull (x#(swap_with_previous xs op)) (Suc n) op) = op \<Longrightarrow> ?case"
    by smt

  from pull_head_to_head_does_nothing
  have ONE:"op \<in> set xs \<and> x = op \<Longrightarrow> hd (pull (swap_with_previous xs op) n op) = op \<Longrightarrow> hd (pull (x#(swap_with_previous xs op)) (Suc n) op) = op"
    by (metis list.sel(1))

  have TWO: "op \<in> set xs \<and> x \<noteq> op \<and> hd xs = op \<Longrightarrow> hd (pull (swap_with_previous xs op) n op) = op \<Longrightarrow> hd (pull (x#(swap_with_previous xs op)) (Suc n) op) = op"
    by (metis  Suc.hyps Suc.prems(1) swap_head_with_prev)

  have "op \<in> set xs \<and> x \<noteq> op \<and> hd xs \<noteq> op \<Longrightarrow> x # (swap_with_previous xs op) = swap_with_previous (x#xs) op" using head_concat_swap by fastforce

  from this and 3 have 4: "op \<in> set xs \<and> x \<noteq> op \<and> hd xs \<noteq> op \<Longrightarrow> hd (pull (swap_with_previous xs op) n op) = op
                                 \<Longrightarrow> hd (pull ((x#swap_with_previous xs op)) (Suc n) op) = op \<Longrightarrow> ?case" by simp

  from local.Suc have "op \<in> set (swap_with_previous xs op) \<and> x \<noteq> op \<and> hd (swap_with_previous xs op) \<noteq> op
                           \<Longrightarrow> hd (pull (swap_with_previous xs op) n op) = op \<Longrightarrow> hd (pull (x#(swap_with_previous xs op)) (Suc n) op) = op" by simp

  from this show ?case using "3" Suc.hyps Suc.prems(1) Suc.prems(2) set_swap_eq by force
qed


lemma list_20: "op \<in> set pi \<Longrightarrow> hd (pull pi (length pi) op) = op"
proof (induct pi)
  case Nil
  then show ?case by simp
next
  case (Cons x xs)
  then have def: "pull (x#xs) (length (x#xs)) op = pull (swap_with_previous (x#xs) op) (length xs) op " by auto
  from this and local.Cons have 1: "x = op \<and> xs = [] \<Longrightarrow> ?case" by simp
  also have 2: "x \<noteq> op \<and> xs = [] \<Longrightarrow> ?case" using Cons.prems by auto
  from 1 and 2 have one: "xs = [] \<Longrightarrow> ?case" by auto

  from def have "x = op \<and> xs \<noteq> [] \<Longrightarrow> pull (x#xs) (length (x#xs)) op = pull (x#xs) (length xs) op" by (metis list.exhaust swap_with_previous.simps(3))
  from this and local.Cons have "x = op \<and> xs \<noteq> [] \<Longrightarrow> ?case" by (metis list.sel(1) pull_head_to_head_does_nothing)

  from local.Cons have "x \<noteq> op \<and> xs \<noteq> [] \<Longrightarrow> ?case" using pull_suc_n by fastforce
  from this show ?case by (metis length_Cons list.sel(1) one  pull_head_to_head_does_nothing)
qed

(* </SWAP WITH PREVIOUS & PULL> *)

(* <PLAN SWP> *)

lemma plan_0:
  fixes T :: "'a ts" and op :: "'a operator" and op' :: "'a operator" and s :: "'a state"
  assumes sInT: "s \<in> states T" and
          validT: "valid_ts T" and
          validOp: "valid_operator T op" and
          validOp':  "valid_operator T op'" and
          independent: "independent T op op'" and
          applicableOpS: "applicable op s" and
          applicableOp'S: "applicable op' s" and
          planOpOp': "plan T s [op, op']"
        shows "plan T s [op', op]"
proof -
  from planOpOp' obtain goal :: "'a state" where "goal \<in> goals T \<and> is_path_from_to T [op, op'] s goal" by (meson plan_def)
  from this show ?thesis using applicableOp'S independent plan_def sInT validOp validOp' validT by (simp add: plan_def enables_def independent_def non_conflict_def)
qed


lemma path_independence_1: "path_independence T pi op \<Longrightarrow> \<forall>op' \<in> set pi. independent T op op' "
proof (induct pi)
  case Nil
  then show ?case by simp
next
  case (Cons a pi)
  then show ?case
    using independent_symmetric by auto
qed


lemma plan_1: "optimal_plan T s pi \<Longrightarrow> plan T s pi"
  by (simp add: optimal_plan_def)


lemma plan_2:
fixes T :: "'a ts" and op :: "'a operator" and op' :: "'a operator" and s :: "'a state"
  assumes sInT: "s \<in> states T" and
          validT: "valid_ts T" and
          validOp: "valid_operator T op" and
          validOp':  "valid_operator T op'" and
          independent: "independent T op op'" and
          applicableOp'S: "applicable op' s" and
          optimalPlanOpOp': "optimal_plan T s [op, op']"
        shows "optimal_plan T s [op', op]"
proof -
  from optimalPlanOpOp' have "plan T s [op, op']" by (simp add: plan_1)
  from this and plan_0 have 1: "plan T s [op', op]" using applicableOp'S independent independent_def plan_def using validT by fastforce
  from this have "path_cost [op', op] = path_cost [op, op']" by auto
  from this have "path_cost [op', op] = Min{path_cost pi' | pi'. plan T s pi'}" using optimalPlanOpOp' optimal_plan_def by force
  from this and 1 show ?thesis by (simp add: optimal_plan_def)
qed


lemma plan_3:
  fixes T :: "'a ts" and op :: "'a operator" and op' :: "'a operator" and s :: "'a state"
  assumes validTS: "valid_ts T" and
          independent: "independent T op op'" and
          plan: "plan T s [op, op']" and
          sInT: "s \<in> states T" and
          applicableOp's: "applicable op' s"
        shows "plan T s [op', op]"
  by (meson applicableOp's independent plan plan_0 plan_def validTS valid_path.simps(2))


lemma plan_4:
  fixes T :: "'a ts" and ys :: "'a path" and s :: "'a state"
  shows "valid_path T (xs @ ys) s \<Longrightarrow> is_path_from_to T xs s (path_effect T xs s)"
proof(induct xs arbitrary: s)
  case Nil
  then show ?case by (simp add: path_effect_1)
next
  case (Cons x xs)
  then have "applicable x s \<Longrightarrow> valid_path T xs (effect T x s) \<Longrightarrow> is_path_from_to T (x#xs) s (path_effect T (x#xs) s)"  by (simp add: path_effect_1)
  then show ?case using Cons.hyps Cons.prems by auto
qed


lemma plan_5:
fixes T :: "'a ts" and ys :: "'a path" and s :: "'a state"
  shows "valid_path T (xs @ ys) s \<Longrightarrow> valid_path T xs s"
proof(induct xs arbitrary: s)
  case Nil
  then show ?case by (simp add: path_effect_1)
next
  case (Cons x xs)
  then have "applicable x s \<Longrightarrow> valid_path T xs (effect T x s) \<Longrightarrow> is_path_from_to T (x#xs) s (path_effect T (x#xs) s)"  by (simp add: path_effect_1)
  then show ?case using Cons.hyps Cons.prems by auto
qed


lemma plan_6:
fixes T :: "'a ts" and ys :: "'a path" and s :: "'a state"
  shows "valid_path T (xs @ ys) s \<Longrightarrow> is_path_from_to T ys (path_effect T xs s) (path_effect T ys (path_effect T xs s))"
proof(induct xs arbitrary: s)
  case Nil
  then show ?case by (simp add: path_effect_1)
next
  case (Cons x xs)
  then have "applicable x s \<Longrightarrow> valid_path T xs (effect T x s) \<Longrightarrow> is_path_from_to T (x#xs) s (path_effect T (x#xs) s)"  by (simp add: path_effect_1)
  then show ?case using Cons.hyps Cons.prems by auto
qed


lemma plan_7:
fixes T :: "'a ts" and ys :: "'a path" and s :: "'a state"
  shows "valid_path T (xs @ ys) s \<Longrightarrow> valid_path T ys (path_effect T xs s)"
proof(induct xs arbitrary: s)
  case Nil
  then show ?case by (simp add: path_effect_1)
next
  case (Cons x xs)
  then have "applicable x s \<Longrightarrow> valid_path T xs (effect T x s) \<Longrightarrow> is_path_from_to T (x#xs) s (path_effect T (x#xs) s)"  by (simp add: path_effect_1)
  then show ?case using Cons.hyps Cons.prems by auto
qed


lemma plan_8:
  fixes T :: "'a ts" and x :: "'a operator" and s :: "'a state"
  assumes validTS: "valid_ts T" and validX: "valid_operator T x" and appX: "applicable x s" and sInT: "s \<in> states T"
  shows "effect T x s \<in> states T"
proof -
  have "(\<exists> s' \<in> states T. \<exists>t\<in>x. s = src t \<and> s' = dst t)" by (metis appX applicable_def in_mono validTS validX valid_operator_def valid_ts_def)

  from this obtain s' :: "'a state" where "s' \<in> states T \<and> (\<exists> t \<in> x. s = src t \<and> s' = dst t)" by auto
  from this and someI_ex have "s' = (SOME s' :: 'a state. s' \<in> states T \<and> (\<exists> t \<in> x. s = src t \<and> s' = dst t))" by (smt validX valid_operator_def)
  from this have "s' = effect T x s" by (simp add: effect_def)
  show ?thesis using \<open>s' = effect T x s\<close> \<open>s' \<in> states T \<and> (\<exists>t\<in>x. s = src t \<and> s' = dst t)\<close> by auto
qed


lemma plan_9:
  assumes "valid_ts T"
  shows "plan T s (x#xs) \<Longrightarrow> plan T (effect T x s) (xs)"
proof (induct xs)
  case Nil
  have "effect T x s \<in> states T" using Nil.prems assms path_effect_0 plan_def valid_ts_def by fastforce
  then show ?case by (meson Nil.prems is_path_from_to.simps(2) plan_def valid_path.simps(2))
next
  case (Cons x xs)
  have "plan T s (x#xs) \<Longrightarrow> valid_path T (x#xs) s" by (simp add: plan_def)
  have "valid_path T (x#xs) s \<Longrightarrow> valid_operator T x" by simp
  from this show ?case by (meson Cons.prems assms is_path_from_to.simps(2) plan_8 plan_def valid_path.simps(2))
qed


lemma plan_10:
  fixes T :: "'a ts" and ys :: "'a path" and s :: "'a state"
  assumes validTS: "valid_ts T"
  shows "plan T s (xs @ ys) \<Longrightarrow> plan T (path_effect T xs s) ys"
proof(induct xs arbitrary: s)
  case Nil
  then show ?case by simp
next
  case (Cons x xs)
  then have "plan T (effect T x s) (xs @ ys) \<Longrightarrow>  plan T (path_effect T (x#xs) s) ys \<Longrightarrow> ?case" by simp
  from this have 1: "plan T (effect T x s) (xs @ ys) \<Longrightarrow>  plan T (path_effect T xs (effect T x s)) ys \<Longrightarrow> ?case" by simp
  from local.Cons have 2: "plan T (effect T x s) (xs @ ys) \<Longrightarrow>  plan T (path_effect T xs (effect T x s)) ys" by auto
  have "plan T (path_effect T xs (effect T x s)) ys \<Longrightarrow> plan T (path_effect T (x # xs) s) ys" by simp
  from this have  "plan T (effect T x s) (xs @ ys) \<Longrightarrow> ?case" by (simp add: "2")
  from this show ?case using Cons.prems plan_9 validTS by fastforce
qed


lemma plan_11:
  fixes T :: "'a ts" and op :: "'a operator" and x :: "'a operator" and q :: "'a operator" and s :: "'a state"
  assumes indepXop: "independent T x op" and indepQop: "independent T q op" and planXqOp: "plan T s [x,q,op]" and appOps:"applicable op s" and validTS: "valid_ts T"
  shows "plan T s [x,op,q]"
proof -
  have "plan T (effect T x s) [op,q] \<Longrightarrow> ?thesis" by (meson is_path_from_to.simps(2) planXqOp plan_def valid_path.simps(2))
  from this show ?thesis by (meson appOps enables_def indepQop indepXop independent_def is_path_from_to.simps(2) planXqOp plan_3 plan_9 plan_def validTS)
qed


lemma list_7: "path_cost xs + path_cost ys = path_cost (xs@ys)"
proof (induct xs)
  case Nil
  then show ?case by simp
next
  case (Cons a xs)
  then show ?case by simp
qed


lemma list_8: "path_cost pi = path_cost (swap_with_previous pi op)"
proof (induct pi)
  case Nil
  then show ?case by simp
next
  case (Cons a pi)
  have 1: "pi = [] \<Longrightarrow> ?case" by simp
  have 2: "path_cost (a#pi) = cost a + path_cost pi" by simp
  from this have b:"a = op \<Longrightarrow> path_cost (swap_with_previous (a#pi) op) = cost a + path_cost (swap_with_previous pi op)" by (smt Cons.hyps list.inject swap_with_previous.elims)
  have a:"a \<noteq> op \<Longrightarrow> hd pi \<noteq> op \<Longrightarrow> path_cost (swap_with_previous (a#pi) op) = cost a + path_cost (swap_with_previous pi op)"
    using swap_with_previous.elims apply -
    by (metis head_concat_swap list_6 path_cost.simps(2) set_ConsD)

  have "pi \<noteq> [] \<Longrightarrow> a \<noteq> op \<Longrightarrow> hd pi = op \<Longrightarrow> swap_with_previous (a#pi) op = [hd pi, a] @ tl pi" by (metis append_Cons append_Nil list.exhaust_sel swap_with_previous.simps(3)) 
  from this have "pi \<noteq> [] \<Longrightarrow> a \<noteq> op \<Longrightarrow> hd pi = op \<Longrightarrow> path_cost (swap_with_previous (a#pi) op) = path_cost [hd pi, a] + path_cost (tl pi)" using list_7 apply - by simp
  from this have "pi \<noteq> [] \<Longrightarrow> a \<noteq> op \<Longrightarrow> hd pi = op \<Longrightarrow> path_cost (swap_with_previous (a#pi) op) = cost a + path_cost (hd pi # tl pi)" by simp
  from this have "pi \<noteq> [] \<Longrightarrow> a \<noteq> op \<Longrightarrow> hd pi = op \<Longrightarrow> path_cost (swap_with_previous (a#pi) op) = cost a + path_cost pi" by fastforce
  from this have 3:"pi \<noteq> [] \<Longrightarrow> a \<noteq> op \<Longrightarrow> hd pi = op \<Longrightarrow> ?case" by auto

  show ?case using "1" "2" "3" Cons.hyps a b by linarith
qed


lemma plan_12:
  fixes T :: "'a ts" and op :: "'a operator" and x :: "'a operator" and q :: "'a operator" and s :: "'a state"
  assumes indepXop: "independent T x op" and indepQop: "independent T q op" and planXqOp: "optimal_plan T s [x,q,op]" and appOps:"applicable op s" and validTS: "valid_ts T"
  shows "optimal_plan T s [x,op,q]"
proof -
  have "plan T s [x,q,op]" by (simp add: planXqOp plan_1)
  from this have 1: "plan T s [x,op,q]"  by (simp add: appOps indepQop indepXop plan_11 validTS)
  have "path_cost [x,op,q] = Min{path_cost pi' | pi'. plan T s pi'}" using optimal_plan_def planXqOp by force
  from this show ?thesis by (simp add: "1" optimal_plan_def)
qed


lemma plan_13:
  fixes op :: "'a operator" and s :: "'a state" and T :: "'a ts"
  assumes validTS: "valid_ts T"
  shows "applicable op s \<Longrightarrow> op \<notin> set xs \<Longrightarrow> path_independence T xs op \<Longrightarrow> plan T s (xs @ [op]) \<Longrightarrow> plan T s (swap_with_previous (xs @ [op]) op)"
proof (induct xs arbitrary: s)
  case Nil
  then show ?case by simp
next
  case (Cons x xs)
  {assume "xs = []"
    have 1:"applicable op s \<Longrightarrow> path_independence T (Cons x xs) op \<Longrightarrow> independent T x op" by simp
    have 2:"applicable op s \<Longrightarrow> op \<notin> set (x#xs) \<Longrightarrow> independent T x op \<Longrightarrow> path_independence T (Cons x xs) op" using Cons.prems(3) by auto
    from 1 and 2 have "applicable op s \<Longrightarrow> op \<notin> set (x#xs) \<Longrightarrow> independent T x op \<Longrightarrow> plan T s [x, op] \<Longrightarrow> plan T s [op, x] \<Longrightarrow> ?case" using \<open>xs = []\<close> by auto
    from this have ?case by (metis (no_types, lifting) "1" Cons.prems(1) Cons.prems(2) Cons.prems(3) Cons.prems(4) \<open>xs = []\<close> append_eq_Cons_conv plan_3 plan_def validTS)
  }
  moreover {assume "length xs = 1"
    from this have "\<exists> q :: 'a operator. xs = [q]" by (simp add: list_4)
    from this obtain q :: "'a operator" where "xs = [q]" by auto
    have ?case using Cons.prems(1) Cons.prems(2) Cons.prems(3) Cons.prems(4) \<open>xs = [q]\<close> plan_11 validTS by fastforce
  }

  moreover {assume "length xs > 1"
    have "applicable op s \<Longrightarrow> op \<notin> set (x#xs) \<Longrightarrow> path_independence T (x#xs) op \<Longrightarrow> plan T s ((x#xs) @ [op]) \<Longrightarrow> plan T s (swap_with_previous ((x#xs) @ [op]) op) \<Longrightarrow> ?case" by simp
    from this have "applicable op s \<Longrightarrow> op \<notin> set xs \<Longrightarrow> op \<noteq> x \<Longrightarrow> path_independence T xs op \<Longrightarrow> independent T x op \<Longrightarrow> plan T (effect T x s) (xs @ [op]) \<Longrightarrow> plan T s (swap_with_previous ((x#xs) @ [op]) op) \<Longrightarrow> ?case" by simp
    from this have "applicable op s \<Longrightarrow> op \<notin> set xs \<Longrightarrow> op \<noteq> x \<Longrightarrow> path_independence T xs op \<Longrightarrow> independent T x op \<Longrightarrow> plan T (effect T x s) (xs @ [op]) \<Longrightarrow> plan T s (x # butlast xs @ [op, last xs]) \<Longrightarrow> ?case" by (smt Cons.prems(2) append_Cons append_self_conv2 butlast.simps(2) calculation(1) last_ConsR list.inject list_5)
    from this have "applicable op s \<Longrightarrow> op \<notin> set xs \<Longrightarrow> op \<noteq> x \<Longrightarrow> path_independence T xs op \<Longrightarrow> independent T x op \<Longrightarrow> plan T (effect T x s) (xs @ [op]) \<Longrightarrow> plan T (effect T x s) (butlast xs @ [op, last xs]) \<Longrightarrow> ?case" by (metis Cons.prems(4) append_Cons is_path_from_to.simps(2) plan_def valid_path.simps(2))
    from this have ?case by (metis Cons.hyps Cons.prems(1) Cons.prems(2) Cons.prems(3) Cons.prems(4) append_Cons calculation(1) enables_def independent_def list.set_intros(1) list.set_intros(2) list_5 path_independence.simps(2) plan_9 plan_def validTS valid_path.simps(2))
  }

  ultimately show ?case
    by (meson length_0_conv less_one linorder_neqE_nat)
qed


lemma plan_14: "s \<in> states T \<Longrightarrow> xs \<noteq> [] \<Longrightarrow> valid_ts T \<Longrightarrow> valid_path T xs s \<Longrightarrow> path_effect T xs s \<in> states T"
proof(induct xs arbitrary: s)
  case Nil
  then show ?case by simp
next
  case (Cons x xs)
  {assume "length xs = 1"
   from this obtain q  where "xs = [q]" using list_4 by auto
   from this have "valid_ts T \<Longrightarrow> valid_path T [x, q] s \<Longrightarrow> path_effect T [x, q] s \<in> states T \<Longrightarrow> ?case" by simp
   from this have "valid_ts T \<Longrightarrow> applicable x s \<and> valid_operator T x \<Longrightarrow> path_effect T [x,q] s \<in> states T \<Longrightarrow> ?case" by (simp add: \<open>xs = [q]\<close>)
   have "applicable x s \<and> valid_operator T x \<Longrightarrow> effect T x s \<in> states T" by (simp add: Cons.prems(1) Cons.prems(3) plan_8)
   from this have ?case using Cons.prems(3) Cons.prems(4) \<open>xs = [q]\<close> plan_8 by force
 }

  moreover {assume "length xs > 1"
    have "s \<in> states T \<Longrightarrow> xs \<noteq> [] \<Longrightarrow> valid_ts T \<Longrightarrow> valid_path T xs (effect T x s) \<Longrightarrow> path_effect T (x#xs) s \<in> states T \<Longrightarrow> ?case" by simp
    from this have ?case using Cons.hyps Cons.prems(1) Cons.prems(3) Cons.prems(4) plan_8 by force
  }

  ultimately show ?case using Cons.prems(1) Cons.prems(3) Cons.prems(4) nat_neq_iff plan_8 by fastforce
qed


lemma plan_15:
  fixes s :: "'a state" and T :: "'a ts" and op :: "'a operator" and x :: "'a operator"
  assumes sInT: "s \<in> states T" and validTS: "valid_ts T" and appOp: "applicable op s" and opNotX: "op \<noteq> x" and indep: "independent T x op" and planXOpYs: "plan T s ([x, op] @ ys)"
  shows "plan T s ([op, x] @ ys)"
proof -
  have "applicable x (effect T op s)" by (metis appOp append_Cons enables_def indep independent_def planXOpYs plan_def valid_path.simps(2))
  from this have "effect T x (effect T op s) = effect T op (effect T x s)"  by (metis appOp append_Cons indep independent_def non_conflict_def planXOpYs plan_def valid_path.simps(2))
  from this have eq: "path_effect T [op, x] s = path_effect T [x,op] s" by simp
  obtain s' where "s' = path_effect T [x, op] s" by simp
  from this have "s' \<in> states T" by (meson planXOpYs plan_10 plan_def validTS)
  have "s' = path_effect T [op, x] s" using \<open>s' = path_effect T [x, op] s\<close> eq by auto

  have imp:"(valid_path T [op, x] s \<and> plan T s' ys) \<Longrightarrow> ?thesis" using \<open>effect T x (effect T op s) = effect T op (effect T x s)\<close> path_effect_1 planXOpYs plan_def by fastforce

  have v_path: "valid_path T [op, x] s" using \<open>applicable x (effect T op s)\<close> appOp planXOpYs plan_def by force
  have "plan T s' ys" using \<open>s' = path_effect T [x, op] s\<close> planXOpYs plan_10 validTS by fastforce

  from this and v_path and imp show ?thesis by simp
qed


lemma plan_16:
  fixes op :: "'a operator" and s :: "'a state" and T :: "'a ts" and ys :: "'a path"
  assumes validTS: "valid_ts T"
  shows "s \<in> states T \<Longrightarrow> applicable op s \<Longrightarrow> op \<notin> set xs \<Longrightarrow> path_independence T xs op \<Longrightarrow> plan T s (xs @ [op] @ ys) \<Longrightarrow> plan T s (swap_with_previous (xs @ [op] @ ys) op)"

proof (induct xs arbitrary: s)
  case Nil
  then have "s \<in> states T \<Longrightarrow> applicable op s \<Longrightarrow> path_independence T [] op \<Longrightarrow> plan T s (op # ys) \<Longrightarrow> plan T s (swap_with_previous (op # ys) op) \<Longrightarrow> ?case" by simp
  from this have "s \<in> states T \<Longrightarrow> applicable op s \<Longrightarrow> path_independence T [] op \<Longrightarrow> plan T s (op # ys) \<Longrightarrow> plan T s (op # ys) \<Longrightarrow> ?case" by (metis list.exhaust swap_with_previous.simps(2) swap_with_previous.simps(3))
  from this have "s \<in> states T \<Longrightarrow> applicable op s \<Longrightarrow> path_independence T [] op \<Longrightarrow> plan T s (op # ys) \<Longrightarrow> ?case" by simp
  then show ?case using Nil.prems(1) Nil.prems(2) Nil.prems(5) by auto
next
  case (Cons x xs)
  {assume "xs = []"
    from this have "s \<in> states T \<Longrightarrow> applicable op s \<Longrightarrow> op \<noteq> x \<Longrightarrow> path_independence T [x] op \<Longrightarrow> plan T s ([x, op] @ ys) \<Longrightarrow> plan T s (swap_with_previous ([x,op] @ ys) op) \<Longrightarrow> ?case" by simp
    from this have "s \<in> states T \<Longrightarrow> valid_ts T \<Longrightarrow> applicable op s \<Longrightarrow> op \<noteq> x \<Longrightarrow> independent T x op \<Longrightarrow> plan T s ([x, op] @ ys) \<Longrightarrow> plan T s ([op, x] @ ys) \<Longrightarrow> ?case" by simp
    from this have ?case using Cons.prems(1) Cons.prems(2) Cons.prems(4) Cons.prems(5) \<open>xs = []\<close> path_independence.simps(2) plan_15 validTS by force
  }

  moreover {assume "xs \<noteq> []"
    from this have "s \<in> states T \<Longrightarrow> applicable op s \<Longrightarrow> op \<noteq> x \<Longrightarrow> path_independence T (x#xs) op \<Longrightarrow> plan T s ((x#xs) @ [op] @ ys) \<Longrightarrow> plan T s (swap_with_previous ((x#xs) @ [op] @ ys) op) \<Longrightarrow> ?case" by simp
    from this have "s \<in> states T \<Longrightarrow> applicable op s \<Longrightarrow> op \<noteq> x \<Longrightarrow> independent T op x \<Longrightarrow> path_independence T xs op \<Longrightarrow> plan T s ((x#xs) @ [op] @ ys) \<Longrightarrow> plan T s (swap_with_previous ((x#xs) @ [op] @ ys) op) \<Longrightarrow> ?case" by simp
    from this have "s \<in> states T \<Longrightarrow> applicable op s \<Longrightarrow> op \<noteq> x \<Longrightarrow> independent T op x \<Longrightarrow> path_independence T xs op \<Longrightarrow> applicable x s \<Longrightarrow> plan T (effect T x s) (xs @ [op] @ ys) \<Longrightarrow> plan T s (swap_with_previous ((x#xs) @ [op] @ ys) op) \<Longrightarrow> ?case" by simp
    from this have "s \<in> states T \<Longrightarrow> applicable op s \<Longrightarrow> op \<noteq> x \<Longrightarrow> independent T op x \<Longrightarrow> path_independence T xs op \<Longrightarrow> applicable x s \<Longrightarrow> plan T (effect T x s) (xs @ [op] @ ys) \<Longrightarrow> plan T s (x # swap_with_previous (xs @ [op] @ ys) op) \<Longrightarrow> ?case" by (metis Cons.prems(3) \<open>xs \<noteq> []\<close> list_9)
    from this have "s \<in> states T \<Longrightarrow> applicable op s \<Longrightarrow> op \<noteq> x \<Longrightarrow> independent T op x \<Longrightarrow> path_independence T xs op \<Longrightarrow> applicable x s \<Longrightarrow> plan T (effect T x s) (xs @ [op] @ ys) \<Longrightarrow> applicable x s \<Longrightarrow> plan T (effect T x s) (swap_with_previous (xs @ [op] @ ys) op) \<Longrightarrow> ?case" by (metis Cons.prems(5) append_Cons is_path_from_to.simps(2) plan_def valid_path.simps(2))
    from this have "s \<in> states T \<Longrightarrow> applicable op s \<Longrightarrow> op \<noteq> x \<Longrightarrow> independent T op x \<Longrightarrow> path_independence T xs op \<Longrightarrow> applicable x s \<Longrightarrow> plan T (effect T x s) (xs @ [op] @ ys) \<Longrightarrow> plan T (effect T x s) (swap_with_previous (xs @ [op] @ ys) op) \<Longrightarrow> ?case" by simp
    from this have ?case
      by (metis Cons.hyps Cons.prems(2) Cons.prems(3) Cons.prems(4) Cons.prems(5) append_Cons enables_def independent_def independent_symmetric is_path_from_to.simps(2) list.set_intros(1) list.set_intros(2) path_independence.simps(2) plan_9 plan_def validTS)
  }
  ultimately show ?case by blast
qed

(* </PLAN SWP> *)

(* <COUPLE PLAN SWP WITH THEOREM 1> *)

lemma list_11: "op \<in> set pi \<Longrightarrow> \<exists> xs. \<exists> ys. pi = xs @ [op] @ ys \<and> (\<forall> x \<in> set xs. x \<noteq> op)" by (metis append.left_neutral append_Cons in_set_conv_decomp_first)


lemma list_12: "(\<forall> x \<in> set xs. x \<noteq> op) \<Longrightarrow> \<forall> x \<in> set xs. comes_before x op (xs @ [op] @ ys)"
proof(induct xs)
  case Nil
  then show ?case by simp
next
  case (Cons a xs)
  then show ?case by auto
qed


lemma list_13: "op \<in> set pi \<Longrightarrow>  \<exists> xs. \<exists> ys. (pi = xs @ [op] @ ys \<and> (\<forall> x \<in> set xs. x \<noteq> op \<and> comes_before x op pi))"
  using list_11 list_12 by force


lemma list_14: "\<forall> x \<in> set xs. independent T x op \<Longrightarrow> path_independence T xs op"
proof(induct xs)
  case Nil
  then show ?case by simp
next case (Cons x xs)
  then show ?case by auto
qed


lemma list_15: "\<forall> x \<in> set (xs @ ys). comes_before x op (xs@ys) \<longrightarrow> independent T x op \<Longrightarrow> \<forall> x \<in> set xs. x \<noteq> op \<Longrightarrow> path_independence T xs op"
proof (induct xs)
  case Nil
  then show ?case by simp
next
  case (Cons a xs)
  then have "\<forall> x \<in> set (xs @ ys). comes_before x op (xs@ys) \<longrightarrow> independent T x op \<Longrightarrow> comes_before a op (xs@ys) \<longrightarrow> independent T a op \<Longrightarrow>  \<forall> x \<in> set (a#xs). x \<noteq> op \<Longrightarrow> path_independence T (a#xs) op \<Longrightarrow> ?case" by simp
  then have "\<forall> x \<in> set (xs @ ys). comes_before x op (xs@ys) \<longrightarrow> independent T x op \<Longrightarrow> comes_before a op (xs@ys) \<longrightarrow> independent T a op \<Longrightarrow>  \<forall> x \<in> set xs. x \<noteq> op \<Longrightarrow> a \<noteq> op \<Longrightarrow> path_independence T xs op \<Longrightarrow> independent T a op \<Longrightarrow> ?case" by simp
  from this show ?case by (smt Cons.hyps Cons.prems(1) Cons.prems(2) UnCI Un_iff append_Cons comes_before.simps(2) list.set_intros(1) list.set_intros(2) path_independence.simps(2) set_append set_subset_Cons subset_code(1))
qed


lemma list_16:
  fixes op :: "'a operator" and pi :: "'a path"
  assumes "op \<in> set pi" and "\<forall> x \<in> set pi. comes_before x op pi \<longrightarrow> independent T x op"
  shows "\<exists> xs. \<exists> ys. (pi = xs @ [op] @ ys \<and> (\<forall> x \<in> set xs. x \<noteq> op \<and> comes_before x op pi) \<and> path_independence T xs op)"
proof -
  obtain xs where "\<exists> ys. pi = xs @ [op] @ ys \<and> (\<forall> x \<in> set xs. x \<noteq> op)" using assms(1) list_11 by force
  have "\<forall> x \<in>set xs. comes_before x op pi" using \<open>\<exists>ys. pi = xs @ [op] @ ys \<and> (\<forall>x\<in>set xs. x \<noteq> op)\<close> list_12 by force
  from this have "\<forall> x \<in> set xs. independent T x op" using \<open>\<exists>ys. pi = xs @ [op] @ ys \<and> (\<forall>x\<in>set xs. x \<noteq> op)\<close> assms(2) by auto
  from this have "path_independence T xs op" by (simp add: list_14)
  from this show ?thesis using \<open>\<exists>ys. pi = xs @ [op] @ ys \<and> (\<forall>x\<in>set xs. x \<noteq> op)\<close> \<open>\<forall>x\<in>set xs. comes_before x op pi\<close> by auto
qed


lemma plan_17:
  fixes op :: "'a operator" and s :: "'a state" and T :: "'a ts"
  assumes validTS: "valid_ts T" and
          sInT: "s \<in> states T" and
          appOp: "applicable op s" and
          opInPi: "op \<in> set pi" and
          indep: "\<forall> x \<in> set pi. comes_before x op pi \<longrightarrow> independent T x op" and
          plan: "plan T s pi"
        shows "plan T s (swap_with_previous pi op)"
proof -
  have "op \<in> set pi \<Longrightarrow> \<forall> x \<in> set pi. comes_before x op pi \<longrightarrow> independent T x op \<Longrightarrow> \<exists> xs. \<exists> ys. (pi = xs @ [op] @ ys \<and> (\<forall> x \<in> set xs. x \<noteq> op \<and> comes_before x op pi) \<and> path_independence T xs op)" using list_16 by simp
  from this obtain xs where "\<exists> ys. (pi = xs @ [op] @ ys \<and> (\<forall> x \<in> set xs. x \<noteq> op \<and> comes_before x op pi) \<and> path_independence T xs op)" using indep opInPi by auto
  from this obtain ys where "pi = xs @ [op] @ ys \<and> (\<forall> x \<in> set xs. x \<noteq> op \<and> comes_before x op pi) \<and> path_independence T xs op" by auto
  show ?thesis using \<open>\<exists>ys. pi = xs @ [op] @ ys \<and> (\<forall>x\<in>set xs. x \<noteq> op \<and> comes_before x op pi) \<and> path_independence T xs op\<close> appOp plan plan_16 sInT validTS by fastforce
qed


lemma list_18: "\<forall> x \<in> set pi. comes_before x op pi \<longrightarrow> independent T x op \<Longrightarrow> \<forall> x \<in> set pi. comes_before x op (swap_with_previous pi op) \<longrightarrow> independent T x op" 
proof(induct pi)
  case Nil
  then show ?case by simp
next 
  case (Cons a pi)
  then have "\<forall> x \<in> set (a#pi). comes_before x op (a#pi) \<longrightarrow> independent T x op \<Longrightarrow> \<forall> x \<in> set (a#pi). comes_before x op (swap_with_previous (a#pi) op) \<longrightarrow> independent T x op \<Longrightarrow> ?case" by simp
  from this have "\<forall> x \<in> set pi. comes_before x op (a#pi) \<longrightarrow> independent T x op \<Longrightarrow> comes_before a op (a#pi) \<longrightarrow> independent T a op \<Longrightarrow> \<forall> x \<in> set (a#pi). comes_before x op (swap_with_previous (a#pi) op) \<longrightarrow> independent T x op \<Longrightarrow> ?case" by blast
  from this have "\<forall> x \<in> set pi. comes_before x op (a#pi) \<longrightarrow> independent T x op \<Longrightarrow> independent T a op \<Longrightarrow> \<forall> x \<in> set pi. comes_before x op (swap_with_previous (a#pi) op) \<longrightarrow> independent T x op \<Longrightarrow> comes_before a op (swap_with_previous (a#pi) op) \<longrightarrow> independent T a op \<Longrightarrow> ?case" by fastforce
  from this show ?case by (smt Cons.hyps Cons.prems comes_before.simps(2) list.inject list.set_intros(2) set_ConsD swap_with_previous.elims)
qed


lemma list_19: "\<forall> x \<in> set pi. comes_before x op pi \<longrightarrow> independent T x op \<Longrightarrow> \<forall> x \<in> set pi. comes_before x op (pull pi n op) \<longrightarrow> independent T x op" 
proof(induct n)
  case 0
  then show ?case by simp
next
  case (Suc n)
  then have "\<forall> x \<in> set pi. comes_before x op pi \<longrightarrow> independent T x op \<Longrightarrow> \<forall> x \<in> set pi. comes_before x op (pull pi (Suc n) op) \<longrightarrow> independent T x op \<Longrightarrow> ?case" by simp
  from this have "\<forall> x \<in> set pi. comes_before x op pi \<longrightarrow> independent T x op \<Longrightarrow> \<forall> x \<in> set pi. comes_before x op (swap_with_previous (pull pi n op) op) \<longrightarrow> independent T x op \<Longrightarrow> ?case" by (metis pull.simps(2) pull_and_swap_commute)
  from this show ?case by (metis (no_types, lifting) Suc.hyps Suc.prems list_17 list_18)
qed


lemma plan_18:
  fixes op :: "'a operator" and s :: "'a state" and T :: "'a ts"
  shows "valid_ts T \<Longrightarrow> s \<in> states T \<Longrightarrow> applicable op s \<Longrightarrow> op \<in> set pi \<Longrightarrow>  \<forall> x \<in> set pi. comes_before x op pi \<longrightarrow> independent T x op \<Longrightarrow> plan T s pi \<Longrightarrow> plan T s (pull pi n op)"
proof (induct n)
  case 0
  then show ?case by simp
next
  case (Suc n)
  then have "valid_ts T \<Longrightarrow> s \<in> states T \<Longrightarrow> applicable op s \<Longrightarrow> op \<in> set pi \<Longrightarrow>  \<forall> x \<in> set pi. comes_before x op pi \<longrightarrow> independent T x op \<Longrightarrow> plan T s pi \<Longrightarrow> plan T s (pull pi (Suc n) op) \<Longrightarrow> ?case" by simp
  from this have "valid_ts T \<Longrightarrow> s \<in> states T \<Longrightarrow> applicable op s \<Longrightarrow> op \<in> set pi \<Longrightarrow>  \<forall> x \<in> set pi. comes_before x op pi \<longrightarrow> independent T x op \<Longrightarrow> plan T s pi \<Longrightarrow> plan T s (swap_with_previous (pull pi n op) op) \<Longrightarrow> ?case" by (simp add: pull_and_swap_commute)
  from this have 1:"valid_ts T \<Longrightarrow> s \<in> states T \<Longrightarrow> applicable op s \<Longrightarrow> op \<in> set pi \<Longrightarrow>  \<forall> x \<in> set pi. comes_before x op pi \<longrightarrow> independent T x op \<Longrightarrow> plan T s pi \<Longrightarrow> plan T s (pull pi n op) \<Longrightarrow>  plan T s (swap_with_previous (pull pi n op) op) \<Longrightarrow> ?case" by simp 

  have "valid_ts T \<Longrightarrow> s \<in> states T \<Longrightarrow> applicable op s \<Longrightarrow> op \<in> set(pull pi n op) \<Longrightarrow>  \<forall> x \<in> set (pull pi n op). comes_before x op pi \<longrightarrow> independent T x op \<Longrightarrow> plan T s (pull pi n op) \<Longrightarrow>  plan T s (swap_with_previous (pull pi n op) op)" by (metis (full_types) list_17 list_19 plan_17)
  from this have 2: "valid_ts T \<Longrightarrow> s \<in> states T \<Longrightarrow> applicable op s \<Longrightarrow> op \<in> set pi \<Longrightarrow>  \<forall> x \<in> set pi. comes_before x op pi \<longrightarrow> independent T x op \<Longrightarrow> plan T s (pull pi n op) \<Longrightarrow>  plan T s (swap_with_previous (pull pi n op) op)" using list_17 by force
  from 1 and 2 show ?case by (simp add: Suc.hyps Suc.prems(1) Suc.prems(2) Suc.prems(3) Suc.prems(4) Suc.prems(5) Suc.prems(6)) 
qed


lemma plan_19:
  fixes op :: "'a operator" and s :: "'a state" and T :: "'a ts"
  assumes validTS: "valid_ts T" and
          sInT: "s \<in> states T" and
          appOp: "applicable op s" and
          opInPi: "op \<in> set pi" and
          indep: "\<forall> x \<in> set pi. comes_before x op pi \<longrightarrow> independent T x op" and
          plan: "plan T s pi" and
          piNotNil: "pi \<noteq> []"
        shows "\<exists> ys. plan T s ys \<and> hd ys = op"
proof -
  obtain ys where "ys =  pull pi (length pi) op" by simp
  from this have "plan T s ys" by (simp add: appOp indep opInPi plan plan_18 sInT validTS)
  have "hd ys = op" by (simp add: \<open>ys = pull pi (length pi) op\<close> list_20 opInPi)
  from this show ?thesis using \<open>plan T s ys\<close> by auto
qed


lemma list_21: "path_cost pi = path_cost (pull pi n op)"
proof(induct n)
  case 0
  then show ?case by simp
next
  case (Suc n)
  then show ?case by (simp add: list_8 pull_and_swap_commute)
qed

(* </COUPLE PLAN SWP WITH THEOREM 1> *)

(* <OPTIMAL PLAN SWP> *)

lemma list_22: "optimal_plan T s pi \<Longrightarrow> plan T s pi' \<Longrightarrow> path_cost pi' = path_cost pi \<Longrightarrow> optimal_plan T s pi'"
  by (simp add: optimal_plan_def)


lemma plan_16_dual:
  fixes op :: "'a operator" and s :: "'a state" and T :: "'a ts" and ys :: "'a path"
  assumes validTS: "valid_ts T" and
          sInT: "s \<in> states T" and
          appOp: "applicable op s" and 
          opNotInXs: "op \<notin> set xs" and
          indep: "path_independence T xs op" and
          optimal:"optimal_plan T s (xs @ [op] @ ys)"
        shows "optimal_plan T s (swap_with_previous (xs @ [op] @ ys) op)"
proof -
  have "plan T s (xs @ [op] @ ys)" by (meson optimal plan_1)
  from this have plan:"plan T s (swap_with_previous (xs @ [op] @ ys) op)" by (meson appOp indep opNotInXs plan_16 sInT validTS)
  have "path_cost (xs @ [op] @ ys) = path_cost (swap_with_previous (xs @ [op] @ ys) op)" by (simp add: list_8)
  from this show ?thesis using list_22 optimal plan by fastforce
qed


lemma plan_17_dual:
  fixes op :: "'a operator" and s :: "'a state" and T :: "'a ts"
  assumes validTS: "valid_ts T" and
          sInT: "s \<in> states T" and
          appOp: "applicable op s" and
          opInPi: "op \<in> set pi" and
          indep: "\<forall> x \<in> set pi. comes_before x op pi \<longrightarrow> independent T x op" and
          plan: "optimal_plan T s pi"
        shows "optimal_plan T s (swap_with_previous pi op)"
  by (metis appOp indep list_22 list_8 opInPi plan plan_1 plan_17 sInT validTS)


lemma plan_18_dual:
fixes op :: "'a operator" and s :: "'a state" and T :: "'a ts"
  assumes validTS: "valid_ts T" and
          sInT: "s \<in> states T" and
          appOp: "applicable op s" and
          opInPi: "op \<in> set pi" and
          indep: "\<forall> x \<in> set pi. comes_before x op pi \<longrightarrow> independent T x op" and
          plan: "optimal_plan T s pi"
        shows "optimal_plan T s (pull pi n op)"
  by (metis appOp indep list_21 list_22 opInPi plan plan_1 plan_18 sInT validTS)


lemma plan_19_dual:
  fixes op :: "'a operator" and s :: "'a state" and T :: "'a ts"
  assumes validTS: "valid_ts T" and
          sInT: "s \<in> states T" and
          appOp: "applicable op s" and
          opInPi: "op \<in> set pi" and
          indep: "\<forall> x \<in> set pi. comes_before x op pi \<longrightarrow> independent T x op" and
          plan: "optimal_plan T s pi" and
          piNotNil: "pi \<noteq> []"
        shows "\<exists> ys. optimal_plan T s ys \<and> hd ys = op"
proof -
  obtain ys where "ys =  pull pi (length pi) op" by simp
  from this have "optimal_plan T s ys" by (simp add: appOp indep opInPi plan plan_18_dual sInT validTS)
  have "hd ys = op" by (simp add: \<open>ys = pull pi (length pi) op\<close> list_20 opInPi)
  from this show ?thesis using \<open>optimal_plan T s ys\<close> by auto
qed

(* </OPTIMAL PLAN SWP> *)

theorem MAIN_THEOREM:
  fixes T :: "'a ts" and s :: "'a state" and SSS :: "'a operator set"
  assumes validTS : "valid_ts T" and
          strStbSet : "strong_stubborn_set T SSS s" and
          sInT : "s \<in> states T" and
          solvable : " \<exists> pi :: 'a path. optimal_plan T s pi \<and> pi \<noteq> []"
  shows "\<exists> pi' :: 'a path. hd pi' \<in> SSS \<and> optimal_plan T s pi'"
proof -

  (* obtain some optimal path pi *)
  from solvable obtain pi :: "'a path" where "optimal_plan T s pi \<and> pi \<noteq> []" by auto

  (* obtain the goal state that pi leads to *)
  from this and optimal_plan_def and plan_def have "\<exists> goal \<in> goals T. is_path_from_to T pi s goal \<and> valid_path T pi s" unfolding optimal_plan_def and plan_def by simp
  from this obtain goal :: "'a state" where "goal \<in> goals T \<and> is_path_from_to T pi s goal \<and> valid_path T pi s" by auto

  (* operators in SSS and pi are valid *)
  from strStbSet have validOperatorsInSSS: "\<forall> op \<in> SSS. valid_operator T op" unfolding strong_stubborn_set_def by simp
  from this and sInT and lemma4 and \<open>optimal_plan T s pi \<and> pi \<noteq> []\<close> have validOperatorsInPi: "\<forall> op \<in> set pi. valid_operator T op" by meson

  (* obtain disjunctive landmark L *)
  from strong_stubborn_set_def and strStbSet have  "\<exists> L:: 'a operator set. disjunctive_landmark T L s \<and> L \<subseteq> SSS" by fastforce
  from this obtain L :: "'a operator set" where "disjunctive_landmark T L s \<and> L \<subseteq> SSS" by auto

  (* obtain a valid operator that is in both SSS and pi that is the first occurrence of pi and is also in SSS *)
  from this and validTS and \<open>optimal_plan T s pi \<and> pi \<noteq> []\<close> and lemma1 have interNotEmpty: "SSS \<inter> (set pi) \<noteq> {}" by fastforce
  from this and validOperatorsInPi have validOperatorsInIntersection: "\<forall> op \<in> SSS \<inter> set pi. valid_operator T op" by blast
  from this have "\<forall> op \<in> SSS \<inter> set pi. op \<noteq> {}" using valid_operator_def by auto
  from lemma2 and interNotEmpty and this have "first_occurrence pi SSS \<in> SSS \<inter> set pi" using \<open>optimal_plan T s pi \<and> pi \<noteq> []\<close> by blast
  from this and validOperatorsInIntersection obtain op :: "'a operator" where "op = first_occurrence pi SSS \<and> valid_operator T op \<and> op \<in> SSS \<inter> set pi" by blast

  (* op cannot be s.t. it is not applicable in s *)
  from this have opInPi: "op \<in> set pi" by auto
  from \<open>disjunctive_landmark T L s \<and> L \<subseteq> SSS\<close> and strStbSet and \<open>op = first_occurrence pi SSS \<and> valid_operator T op \<and> op \<in> SSS \<inter> set pi\<close>  have "\<not> applicable op s \<Longrightarrow> necessary_enabling_set T SSS s op" by (metis Diff_insert_absorb inf.cobounded1 not_applicable_in_SSS subset_Diff_insert)
  from this have "\<not> applicable op s \<Longrightarrow> \<exists> op' \<in> SSS. comes_before op' op pi \<and> op' \<in> set pi" by (meson \<open>goal \<in> goals T \<and> is_path_from_to T pi s goal \<and> valid_path T pi s\<close> necessary_enabling_set_def opInPi)
  from this have "\<not> applicable op s \<Longrightarrow> \<exists> op' \<in> SSS. op \<in> SSS \<and>  op' \<in> set pi \<and> op \<in> set pi \<and> valid_operator T op \<and> valid_operator T op' \<and> op = first_occurrence pi SSS \<and> comes_before op' op pi " using \<open>op = first_occurrence pi SSS \<and> valid_operator T op \<and> op \<in> SSS \<inter> set pi\<close> using validOperatorsInSSS by auto
  from this and lemma6 have "\<not> applicable op s \<Longrightarrow> False" by auto

  (* thus op is applicable in s *)
  from this have applicable: "applicable op s" by auto

  (* op is cannot be s.t. it is dependent from any of its previous operators in pi *)
  from validTS and strStbSet and sInT and \<open>optimal_plan T s pi \<and> pi \<noteq> []\<close> and \<open>op = first_occurrence pi SSS \<and> valid_operator T op \<and> op \<in> SSS \<inter> set pi\<close> and  this and lemma7 have "(op' \<in> set pi \<and> comes_before op' op pi \<and> \<not> independent T op op') \<Longrightarrow> False" by fastforce

  (* thus op is independent from the operators that come before it *)
  from this have "op' \<in> set pi \<and> comes_before op' op pi \<Longrightarrow> independent T op op'" by auto

  (* independent operators commute and preserve optimality, as long as the last one is applicable in s *)
  from this have "\<exists> ys. optimal_plan T s ys \<and> hd ys = op" by (smt IntD1 \<open>op = first_occurrence pi SSS \<and> valid_operator T op \<and> op \<in> SSS \<inter> set pi\<close> \<open>optimal_plan T s pi \<and> pi \<noteq> []\<close> applicable applicable_in_SSS independent_symmetric lemma6 opInPi plan_19_dual sInT strStbSet validOperatorsInPi validTS)

  from this show ?thesis using \<open>op = first_occurrence pi SSS \<and> valid_operator T op \<and> op \<in> SSS \<inter> set pi\<close> by blast
qed
end
